@@ -1,6 +1,6 @@
 /*
   xrus - keyboard switcher/indicator and autolock.
-  Copyright (c) 1995-2000 Alexander V. Lukyanov
+  Copyright (c) 1995-2001 Alexander V. Lukyanov
   This is free software with no warranty.
   See COPYING for details.
 */
@@ -18,6 +18,7 @@
 #include <ctype.h>
 #include <sys/types.h>
 #include <signal.h>
+#include <fcntl.h>
 #ifdef HAVE_UNISTD_H
 # include <unistd.h>
 #endif
@@ -187,6 +188,9 @@ XtResource        resources[]=
    {  "occupyAllDesks","OccupyAllDesks",XtRBoolean, sizeof(Boolean),
             XtOffsetOf(XrusRec,occupyAllDesks),
             XtRImmediate,  (XtPointer)True      },
+   {  "keyLog",   "KeyLog",   XtRString,  sizeof(String),
+            XtOffsetOf(XrusRec,keylog_file),
+            XtRString,  (XtPointer)0            },
 };
 
 XrmOptionDescRec  options[]=
@@ -207,9 +211,10 @@ XrmOptionDescRec  options[]=
    {  "+nofork",    "*noFork",     XrmoptionNoArg,   "false"},
    {  "-fork",      "*noFork",     XrmoptionNoArg,   "false"},
    {  "+fork",      "*noFork",     XrmoptionNoArg,   "true" },
-   {  "-wmicon",    "*wmIcon",     XrmoptionNoArg,   "true"},
-   {  "-wmaker",    "*wMaker",     XrmoptionNoArg,   "true"},
-   {  "-perwindow", "*perWindow",  XrmoptionNoArg,   "true"},
+   {  "-wmicon",    "*wmIcon",     XrmoptionNoArg,   "true" },
+   {  "-wmaker",    "*wMaker",     XrmoptionNoArg,   "true" },
+   {  "-perwindow", "*perWindow",  XrmoptionNoArg,   "true" },
+   {  "-keylog",    "*keyLog",     XrmoptionSepArg,  NULL   },
 };
 
 Display        *disp;
@@ -261,6 +266,10 @@ static Bool   pushed=False;
 
 static int xrus_check=1;
 
+static int keylog_fd=-1;
+static int keylog_pos=0;
+static time_t keylog_time=0;
+
 void  SetAlarm();
 
 int   (*old_error_handler)(Display *d,XErrorEvent *ev);
@@ -402,12 +411,14 @@ static int GetKbdState(Window w)
    return -1;
 }
 
+static XtIntervalId SetTitleTimeout=-1;
 static void SetTitleIndicatorOnTimeout(XtPointer closure,XtIntervalId *id)
 {
    Window w=(Window)closure;
    int mode=GetKbdState(w);
    if(mode!=-1)
       SetTitleIndicator(w,mode);
+   SetTitleTimeout=-1;
 }
 
 static Bool window_is_top_level(Window w)
@@ -1017,6 +1028,63 @@ void  MappingNotifyTimeoutHandler(XtPointer closure,XtIntervalId *id)
       QueryFocus();
 }
 
+void LogKeyStr(const char *string,int len)
+{
+   const char *nl;
+
+   if(len==-1)
+      len=strlen(string);
+
+   nl=memchr(string,'\n',len);
+   if(nl)
+      keylog_pos=len-(nl-string+1);
+   else
+      keylog_pos+=len;
+
+   if(keylog_pos>75)
+   {
+      write(keylog_fd,"\n",1);
+      keylog_pos=nl?len-(nl-string+1):len;
+   }
+
+   write(keylog_fd,string,len);
+}
+
+void  LogKey(KeySym ks,char *string,int len)
+{
+   time_t t;
+
+   if(keylog_fd==-1)
+      return;
+
+   if(len>0)
+   {
+      if(string[0]=='\r')
+      {
+         strcpy(string,"<CR>\n");
+         len=-1;
+      }
+      else if((unsigned char)string[0] < 32)
+      {
+         sprintf(string,"^%c",string[0]+'@');
+         len=2;
+      }
+   }
+   else
+   {
+      sprintf(string,"<%.60s>",XKeysymToString(ks));
+      len=-1;
+   }
+   if(time(&t)-keylog_time>=60)
+   {
+      char ts[64];
+      strftime(ts,sizeof(ts),"<%Y-%m-%d %H:%M:%S>",localtime(&t));
+      LogKeyStr(ts,-1);
+      keylog_time=t;
+   }
+   LogKeyStr(string,len);
+}
+
 void  MainLoop(void)
 {
    XEvent   ev;
@@ -1040,6 +1108,12 @@ void  MainLoop(void)
 
       KeepTrackOfKeyboard(&ev);
 
+      if(NewMode!=Mode)
+      {
+         FixNewMode();
+         SwitchKeyboard(NewMode);
+      }
+
       switch(ev.type)
       {
       case(MapNotify):
@@ -1074,6 +1148,8 @@ void  MainLoop(void)
          if(len>0)
             NewMode&=~FOR_ONE;
 
+         LogKey(ks,string,len);
+
          FixNewMode();
          SwitchKeyboard(NewMode);
 
@@ -1138,7 +1214,9 @@ void  MainLoop(void)
          if(AppData.per_window_state && ev.xproperty.atom==XA_WM_NAME
          && ev.xproperty.state==PropertyNewValue)
          {
-            XtAppAddTimeOut(app_context,1000,
+            if(SetTitleTimeout!=-1)
+               XtRemoveTimeOut(SetTitleTimeout);
+            SetTitleTimeout=XtAppAddTimeOut(app_context,1000,
                         (XtTimerCallbackProc)SetTitleIndicatorOnTimeout,
                         (XtPointer)ev.xproperty.window);
          }
@@ -1311,6 +1389,13 @@ void  hook_signals(void)
    act.sa_handler=sig_chld;
    act.sa_flags=SA_RESTART|SA_NOCLDSTOP;
    sigaction(SIGCHLD,&act,NULL);
+
+   act.sa_handler=ToLatKeysFire;
+   act.sa_flags=0;
+   sigaction(SIGUSR1,&act,NULL);
+   act.sa_handler=ToRusKeysFire;
+   act.sa_flags=0;
+   sigaction(SIGUSR2,&act,NULL);
 }
 
 int   main(int argc,char **argv)
@@ -1341,7 +1426,7 @@ int   main(int argc,char **argv)
       {
          printf("Xrus(keyboard switcher) version " VERSION "\n"
                 "\n"
-                "Copyright (c) 1995-2000 Alexander V. Lukyanov (lav@yars.free.net)\n"
+                "Copyright (c) 1995-2001 Alexander V. Lukyanov (lav@yars.free.net)\n"
                 "This is free software that gives you freedom to use, modify and distribute\n"
                 "it under certain conditions, see COPYING (GNU GPL) for details.\n"
                 "There is ABSOLUTELY NO WARRANTY, use at your own risk.\n");
@@ -1468,6 +1553,13 @@ int   main(int argc,char **argv)
    ParseKeyCombination(AppData.switchForOneKeys,&SwitchForOneKeys,err);
    PrintParseErrors("switchForOneKeys",err);
 
+   if(AppData.keylog_file)
+   {
+      keylog_fd=open(AppData.keylog_file,O_WRONLY|O_APPEND|O_CREAT,0600);
+      if(keylog_fd==-1)
+         perror(AppData.keylog_file);
+   }
+
 #if !defined(DEBUG) && !defined(DONT_FORK)
    if(!AppData.noFork && !AppData.wmaker_icon)
    {
